"""Constants according to epc qr code specification."""

ALLOWED_KEYS = ["beneficiary", "iban", "amount", "remittance"]

ROW_MAPPING = {
    1: "bcd",
    2: "version",
    3: "encoding",
    4: "identification_code",
    5: "bic",
    6: "beneficiary",
    7: "iban",
    8: "amount",
    9: "purpose",
    10: "remittance_structured",
    11: "remittance_unstructured",
    12: "originator_information",
}

ENCODINGS = {
    1: "UTF-8",
    2: "ISO-8859-1",
    3: "ISO-8859-2",
    4: "ISO-8859-4",
    5: "ISO-8859-5",
    6: "ISO-8859-7",
    7: "ISO-8859-10",
    8: "ISO-8859-15",
}



"""Implements the command line interface (CLI)."""

import typer

app = typer.Typer()

from py_epc_qr import __version__
from py_epc_qr.checks import (
    check_amount,
    check_beneficiary,
    check_iban,
    check_remittance_unstructured,
    validate_prompt,
)
from py_epc_qr.transaction import consumer_epc_qr
from PIL import Image


@app.command()
def create(
    out: str = typer.Option(
        default="qr.png",
        help="name of generated qr png file",
    ),
    from_yaml: str = typer.Option(
        default="",
        help="specify yaml file from which to create qr",
    ),
):
    """
    Create EPC-compliant QR code for IBAN-based wire transfer within European economic area.
    """

    if from_yaml:
        typer.echo("creating qr code from yaml...")
        epc = consumer_epc_qr.from_yaml(from_yaml)
    else:
        beneficiary = typer.prompt("Enter the beneficiary", type=str)
        if not validate_prompt(check_beneficiary(beneficiary)):
            typer.echo("The beneficiary is not valid.")
            raise typer.Exit(code=1)
        iban = typer.prompt("Enter the IBAN", type=str)
        if not validate_prompt(check_iban(iban)):
            typer.echo("The IBAN appears incorrect.")
            raise typer.Exit(code=1)
        amount = typer.prompt("Enter the amount", type=float)
        if not validate_prompt(check_amount(amount)):
            typer.echo("The amount appears incorrect (must be float).")
            raise typer.Exit(code=1)
        remittance = typer.prompt("Enter reason for payment", type=str)
        if not validate_prompt(check_remittance_unstructured(remittance)):
            typer.echo("The value for the remittance appears incorrect.")
            raise typer.Exit(code=1)
        epc = consumer_epc_qr(beneficiary, iban, amount, remittance)

    epc.to_qr(out)
    typer.echo(f"🎉🎉🎉 You may view your png {out} 🎉🎉🎉")


@app.command()
def version():
    """
    Show version and exit.
    """
    typer.echo(f"py-epc-qr v{__version__}")
    raise typer.Exit()


@app.callback()
def main():
    """
    Create EPC-compliant QR codes for wire transfers.
    """



"""Several functions that validate the epc qr code format."""

from collections import namedtuple

check = namedtuple("Check", ["valid", "error"])


def check_version(value: str, bic: str) -> tuple:
    """
    Checks whether the version entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    if value not in (valid_versions := ["001", "002"]):
        return check(
            False,
            ValueError(f"invalid version `{value}` (choose from {valid_versions}"),
        )
    if value == "001" and not bic:
        return check(False, AssertionError("version 001 requires a BIC"))
    return check(True, None)


def check_amount(value: float) -> tuple:
    """
    Checks whether the amount entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    try:
        value = float(value)
    except Exception as e:
        return check(False, ValueError("amount must be convertible to float"))
    if not 0.01 <= value <= 999999999.99:
        return check(False, ValueError(f"the amount {value} is out of bounds"))

    if value != round(value, 2):
        return check(False, ValueError("the amount is not a two-digit decimal number"))
    return check(True, None)


def check_encoding(value: str) -> tuple:
    """
    Checks whether the encoding entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    if not 1 <= int(value) <= 8:
        return check(False, ValueError("encoding must be between 1 and 8"))
    return check(True, None)


def check_beneficiary(value: str) -> tuple:
    """
    Checks whether the beneficiary entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    if not value.replace(" ", "").isalnum():
        return check(False, ValueError("beneficiary is not alphanumeric"))
    if not 1 <= len(value) <= (max_length := 70):
        return check(
            False,
            ValueError(
                f"beneficiary is mandatory, and must not exceed {max_length} characters"
            ),
        )
    return check(True, None)


def check_iban(value: str) -> tuple:
    """
    Checks whether the iban entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    if not value.isalnum():
        return check(False, ValueError("iban is not alphanumeric"))
    country_code = value[0:1]
    check_digits = value[2:3]
    bban = value[4:]
    if not country_code.isalpha():
        return check(False, ValueError("invalid iban country code"))
    if not check_digits.isnumeric():
        return check(False, ValueError("invalid check digits"))
    if len(bban) > 30:
        return check(False, ValueError("bban is too long"))
    return check(True, None)


def check_remittance_unstructured(value: str) -> tuple:
    """
    Checks whether the unstructured remittance entry is valid.
    Returns a namedtuple of kind `check`, which is either `(True, None)` or `(False, AssertionError)`.
    """
    if not value.replace(" ", "").isalnum():
        return check(False, ValueError("unstructered remittance is non alphanumeric"))
    if len(value) > 140:
        return check(
            False, ValueError("unstructured remittance exceeds 140 characters")
        )
    return check(True, None)


def validate(res: namedtuple) -> None:
    """
    Raises the error from the `check` namedtuple, if any.
    """
    if not res.valid and res.error is not None:
        raise res.error


def validate_prompt(res: namedtuple) -> None:
    """
    Raises the boolean from the `check` namedtuple.
    """
    return res.valid



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py_epc_qr']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.0,<10.0.0',
 'PyYAML>=6.0,<7.0',
 'qrcode>=7.3.1,<8.0.0',
 'typer>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['epcqr = py_epc_qr.__main__:main']}

setup_kwargs = {
    'name': 'py-epc-qr',
    'version': '0.1.2.2',
    'description': 'Generate EPC-compatible QR codes for wire transfers',
    'long_description': '# Create QR codes for wire transfers\n\nSick of copy-and-pasting IBANs to forms?\nWhy not just scan a QR code and have your favorite banking app take care of the rest?\n\nWhy not be generous and support wikipedia with 123,45€?\nGrab your phone and scan the image.\n\n![Support Wikipedia with 123,45 €](tests/data/qr_version_002.png "Support Wikipedia with 123,45 €")\n\n[The create QR code complies with the European Payments Council (EPC) Quick Response (QR) code guidelines.](https://en.wikipedia.org/wiki/EPC_QR_code)\n\nDisclaimer: The author of this code has no affiliation with the EPC whatsoever.\nHenceforth, you are welcome to use the code at your own dispense, but any use is at your own (commercial) risk.\n\n## Installation\n\nYou can easily install the Python package via pip.\n\n```python\npip install py-epc-qr\n```\n\n## Usage\n\nYou can use the package as part of your own code or as a standalone command line interface (CLI).\n\n### Code\n\n```python\nfrom py_epc_qr.transaction import consumer_epc_qr\nepc_qr = consumer_epc_qr(\n    beneficiary= "Wikimedia Foerdergesellschaft",\n    iban= "DE33100205000001194700",\n    amount= 123.45,\n    remittance= "Spende fuer Wikipedia"\n    )\nepc_qr.to_qr()\n```\n\n### CLI\n\n<todo>\n\n### From interaction\n\n<todo>\n\n#### From template\n\n<todo>\n\n## Limitations\n\nCurrently, the EPC specifications are implemented only to work with IBAN-based consumer wire transfers within the European Economic Area.',
    'author': 'timueh',
    'author_email': 't.muehlpfordt@mailbox.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/timueh/py-epc-qr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Core functionality that converts epc qr code format to qr code image."""

import qrcode
import yaml

from py_epc_qr.checks import (
    check_amount,
    check_beneficiary,
    check_encoding,
    check_iban,
    check_remittance_unstructured,
    check_version,
    validate,
)
from py_epc_qr.constants import ALLOWED_KEYS, ENCODINGS, ROW_MAPPING


class epc_qr:
    """
    Class containing epc qr code specification and its conversion to image/text.
    """

    def __init__(
        self,
        version: str,
        encoding: int,
        bic: str,
        beneficiary: str,
        iban: str,
        amount: float,
        purpose: str,
        remittance_structured: str,
        remittance_unstructured: str,
        originator_information: str,
    ):
        """Initialize"""
        self.bcd = "BCD"
        self.bic = bic
        self.version = version
        self.encoding = encoding
        self.identification_code = "SCT"
        self.beneficiary = beneficiary
        self.iban = iban
        self.amount = amount
        self.purpose = purpose
        self.remittance_structured = remittance_structured
        self.remittance_unstructured = remittance_unstructured
        self.originator_information = originator_information

    def to_txt(self, file_name: str = "qr_source.txt") -> None:
        """
        Write EPC-compliant string to text file `file_name`
        """
        with open(file_name, "w", encoding=self.resolve_encoding()) as file:
            for key, value in ROW_MAPPING.items():
                file.write(self.__getattribute__(value))
                if key < 12:
                    file.write("\n")

    def to_str(self) -> str:
        """
        Write EPC-compliant string.
        """
        res = ""
        for key, value in ROW_MAPPING.items():
            res += self.__getattribute__(value)
            if key < 12:
                res += "\n"
        return res

    def to_qr(self, file_name: str = "qr.png"):
        """
        Write EPC-compliant string to png image `file_name`
        """
        qr = qrcode.QRCode(
            version=6,
            error_correction=qrcode.constants.ERROR_CORRECT_M,
        )
        qr.add_data(self.to_str())
        qr.make()
        img = qr.make_image()
        img.save(file_name)
        print("created image")

    # Properties of class

    @property
    def version(self) -> str:
        """
        Return EPC version entry.
        """
        return self.__version

    @version.setter
    def version(self, value: str):
        """
        Set EPC version entry.
        """
        validate(check_version(value, self.bic))
        self.__version = value

    @property
    def amount(self) -> str:
        """
        Return EPC amount entry.
        """
        return self.__amount

    @amount.setter
    def amount(self, value):
        """
        Set and validate EPC amount entry.
        """
        validate(check_amount(value))
        self.__amount = "EUR{:.2f}".format(float(value))

    @property
    def encoding(self) -> str:
        """
        Return EPC encoding entry.
        """
        return self.__encoding

    @encoding.setter
    def encoding(self, value) -> None:
        """
        Set and validate EPC encoding entry.
        """
        validate(check_encoding(value))
        self.__encoding = str(value)

    def resolve_encoding(self) -> str:
        """
        Resolve EPC encoding entry.
        """
        return ENCODINGS[int(self.encoding)]

    @property
    def beneficiary(self) -> str:
        """
        Return EPC beneficiary entry.
        """
        return self.__beneficiary

    @beneficiary.setter
    def beneficiary(self, value: str) -> None:
        """
        Set and validate EPC beneficiary entry.
        """
        validate(check_beneficiary(value))
        self.__beneficiary = value

    @property
    def iban(self) -> str:
        """
        Return EPC iban entry.
        """
        return self.__iban

    @iban.setter
    def iban(self, value: str):
        """
        Set and validate EPC iban entry.
        """
        validate(check_iban(value))
        self.__iban = value

    @property
    def remittance_unstructured(self) -> str:
        """
        Return EPC unstructured remittance entry.
        """
        return self.__remittance_unstructured

    @remittance_unstructured.setter
    def remittance_unstructured(self, value) -> None:
        """
        Set and validate EPC unstructured remittance entry.
        """
        validate(check_remittance_unstructured(value))
        self.__remittance_unstructured = value


class consumer_epc_qr(epc_qr):
    """
    Standard consumer EPC QR code for IBAN-based wire transfer within European economic area.
    """

    def __init__(self, beneficiary: str, iban: str, amount: float, remittance: str):
        """Initialize"""
        super().__init__(
            version="002",
            encoding=1,
            bic="",
            beneficiary=beneficiary,
            iban=iban,
            amount=amount,
            purpose="",
            remittance_structured="",
            remittance_unstructured=remittance,
            originator_information="",
        )

    @classmethod
    def from_yaml(cls, file_name: str):
        """Create from yaml file"""
        with open(file_name, "r") as file:
            data = yaml.safe_load(file)
        if not sorted(list(data.keys())) == sorted(ALLOWED_KEYS):
            raise AssertionError(
                f"yaml template has incorrect entries (allowed are {ALLOWED_KEYS})"
            )
        return cls(**data)
